import sys
from sys import version_info

if sys.version_info >= (3, 4, 3): ...  # PYI004 
if sys.version_info < (3, 4, 3): ...  # PYI004 
if sys.version_info == (3, 4, 3): ...  # PYI004 
if sys.version_info != (3, 4, 3): ...  # PYI004 

if sys.version_info[0] == 2: ...
if version_info[0] == 2: ...
if sys.version_info < (3, 5): ...
if version_info >= (3, 5): ...
if sys.version_info[:2] == (2, 7): ...
if sys.version_info[:1] == (2,): ...
if sys.platform == 'linux': ...
