#!/usr/bin/env python3
"""Here's a top-level ß9💣2ℝing that's over theß9💣2ℝ."""


def f1():
    """Here's a ß9💣2ℝing that's also over theß9💣2ℝ."""

    x = 1  # Here's a comment that's over theß9💣2ℝ, but it's not standalone.

    # Here's a standalone comment that's over theß9💣2ℝ.

    x = 2
    # Another standalone that is preceded by a newline and indent toke and is over theß9💣2ℝ.

    print("Here's a string that's over theß9💣2ℝ, but it's not a ß9💣2ℝing.")


"This is also considered a ß9💣2ℝing, and is over theß9💣2ℝ."


def f2():
    """Here's a multi-line ß9💣2ℝing.

    It's over theß9💣2ℝ on this line, which isn't the first line in the ß9💣2ℝing.
    """


def f3():
    """Here's a multi-line ß9💣2ℝing.

    It's over theß9💣2ℝ on this line, which isn't the first line in the ß9💣2ℝing."""
