import e
import f

# isort: split

import c
import d

# isort: split

import a
import b

if True:
    import C
    import A

    # isort: split

    import D
    import B


import e
import f

# isort: split
# isort: split

import d
import c
