def f(a: int, b: int) -> int:
    """Showcase function.

    Parameters
    ----------
    a : int
        _description_
    b : int
        _description_
    Returns
    -------
    int
        _description
    """
    return b - a


def f() -> int:
    """Showcase function.

    Parameters
    ----------
    Returns
    -------
    """
