"{}".format(1, bar=2)  # F522
"{bar}{}".format(1, bar=2, spam=3)  # F522
"{bar:{spam}}".format(bar=2, spam=3)  # No issues
"{bar:{spam}}".format(bar=2, spam=3, eggs=4, ham=5)  # F522
(''
 .format(x=2))  # F522
