from a import aksjdhflsakhdflkjsadlfajkslhfdkjsaldajlahflashdfljahlfksajlhfajfjfsaahflakjslhdfkjalhdskjfa
from a import aksjdhflsakhdflkjsadlfajkslhfdkjsaldajlahflashdfljahlfksajlhfajfjfsaahflakjslhdfkjalhdskjfa, aksjdhflsakhdflkjsadlfajkslhfdkjsaldajlahflashdfljahlfksajlhfajfjfsaahflakjslhdfkjalhdskjfa
from a import aksjdhflsakhdflkjsadlfajkslhfdkjsaldajlahflashdfljahlfksajlhfajfjfsaahflakjslhdfkjalhdskjfa as dfgsdfgsd, aksjdhflsakhdflkjsadlfajkslhfdkjsaldajlahflashdfljahlfksajlhfajfjfsaahflakjslhdfkjalhdskjfa as sdkjflsdjlahlfd

# At the top-level, force one empty line after an import, but allow up to two empty
# lines.
import os
import sys
x = 1

import os
import sys

x = 1

import os
import sys


x = 1

import os
import sys



x = 1


# In a nested scope, force one empty line after an import.
def func():
    import os
    import sys
    x = 1


def func():
    import os
    import sys

    x = 1


def func():
    import os
    import sys


    x = 1


def func():
    import os
    import sys



    x = 1
