# noqa
# noqa # comment
print()  # noqa
print()  # noqa # comment
print()  # noqa  # comment
print()  # noqa comment
print()  # noqa  comment
print(a)  # noqa
print(a)  # noqa # comment
print(a)  # noqa  # comment
print(a)  # noqa comment
print(a)  # noqa  comment

# noqa: E501, F821
# noqa: E501, F821 # comment
print()  # noqa: E501, F821
print()  # noqa: E501, F821 # comment
print()  # noqa: E501, F821  # comment
print()  # noqa: E501, F821 comment
print()  # noqa: E501, F821  comment
print(a)  # noqa: E501, F821
print(a)  # noqa: E501, F821 # comment
print(a)  # noqa: E501, F821  # comment
print(a)  # noqa: E501, F821 comment
print(a)  # noqa: E501, F821  comment
