import mod.BadAllowed as badallowed
import mod.stillBad as stillbad

from mod import BadAllowed as badallowed
from mod import StillBad as stillbad

from mod import BadAllowed as bad_allowed
from mod import StillBad as still_bad
