def not_checked():
    import math


def unconventional():
    import altair
    import matplotlib.pyplot
    import numpy
    import pandas
    import seaborn
    import tkinter
    import networkx


def unconventional_aliases():
    import altair as altr
    import matplotlib.pyplot as plot
    import numpy as nmp
    import pandas as pdas
    import seaborn as sbrn
    import tkinter as tkr
    import networkx as nxy

def conventional_aliases():
    import altair as alt
    import matplotlib.pyplot as plt
    import numpy as np
    import pandas as pd
    import seaborn as sns
    import tkinter as tk
    import networkx as nx
