def func():
    try:
       ...
    except Exception:
       ...
    except OtherException as e:
       ...
    else:
       ...
    finally:
       ...

def func():
    try:
       ...
    except Exception:
       ...

def func():
    try:
       ...
    except Exception:
       ...
    except OtherException as e:
       ...

def func():
    try:
       ...
    except Exception:
       ...
    except OtherException as e:
       ...
    else:
       ...

def func():
    try:
       ...
    finally:
       ...
