def a():
    return

def __init__():
    return

class A:
    def __init__(self):
        return


class B:
    def __init__(self):
        return 3

    def gen(self):
        return 5

class MyClass:

    def __init__(self):
        return 1

class MyClass2:
    """dummy class"""

    def __init__(self):
        return


class MyClass3:
    """dummy class"""

    def __init__(self):
        return None

class MyClass5:
    """dummy class"""

    def __init__(self):
        self.callable = lambda: (yield None)
