import mod.BadAllowed as BADALLOWED
import mod.StillBad as STILLBAD

from mod import BadAllowed as BADALLOWED
from mod import StillBad as STILLBAD

from mod import BadAllowed as BAD_ALLOWED
from mod import StillBad as STILL_BAD
